/******************************************************************************
 * arcball.c                                                                  *
 *                                                                            *
 * Computer Graphics and Visualization - Fall 05 - Project 3                  *
 *                                                                            *
 * Erik Waling          <c03ewg@cs.umu.se>                                    *
 * December 26 - 2005                                                         *
 ******************************************************************************/

#include <stdio.h>
#include <GL/gl.h>
#include <string.h>
#include <math.h>
#include "vectormath.h"
#include "arcball.h"

/* The arcball method that allows a user to rotate an object by 
 * dragging the mouse is described at:
 * http://nehe.gamedev.net/data/lessons/lesson.asp?lesson=48
 */

/* Current rotation. */
static GLfloat thisRotation[3][3] = {
	{1.0, 0.0, 0.0},
	{0.0, 1.0, 0.0},
	{0.0, 0.0, 1.0}
};

/* Total accumulated rotation. */
static GLfloat lastRotation[3][3] = {
	{1.0, 0.0, 0.0},
	{0.0, 1.0, 0.0},
	{0.0, 0.0, 1.0}
};

/* Matrix to give GL. This matrix will contain thisRotation * lastRotation */
static GLfloat transformMatrix[16] = {
	1.0, 0.0, 0.0, 0.0,
	0.0, 1.0, 0.0, 0.0,
	0.0, 0.0, 1.0, 0.0,
	0.0, 0.0, 0.0, 1.0
};

/* Contains the first mapped vector (when the mouse button is pressed down.) */
static GLfloat mouseStartVect[3];

/* Scaling factors that are adjusted when window is resized. */
static GLfloat adjWidth;
static GLfloat adjHeight;

void mapToSphere(GLfloat mouseX, GLfloat mouseY, GLfloat vector[3])
{
	GLfloat sqLength;
	GLfloat normFactor;
	
	mouseX = (mouseX * adjWidth) - 1.0;
	mouseY = 1.0 - (mouseY * adjHeight);

	sqLength = mouseX * mouseX + mouseY * mouseY;

	if (sqLength > 1.0) {
		/* Outside sphere. Map to closest point on sphere. */
		
		normFactor = 1.0 / sqrt(sqLength);
		
		vector[0] = mouseX * normFactor;
		vector[1] = mouseY * normFactor;
		vector[2] = 0.0; /* 0.0 since sphere is centered in origin */
			
	} else {
		/* Inside sphere. */
		vector[0] = mouseX;
		vector[1] = mouseY;
		vector[2] = sqrt(1.0 - sqLength);
	}
}

void rotateArcBall(GLfloat mouseEndX, GLfloat mouseEndY)
{
	/* Called when the mouse is dragged. */
	GLfloat mouseEndVect[3];
	GLfloat perpVect[3];
	GLfloat rotQuat[4] = {0.0, 0.0, 0.0, 0.0};
	
	/* Map end vector to spehere. */
	mapToSphere(mouseEndX, mouseEndY, mouseEndVect);
	
	/* Find the perpendicular vector we want to rotate around. */
	crossProduct(mouseStartVect, mouseEndVect, perpVect);
	
	if (norm(perpVect) > 0.01) {
		/* Create the quaternion that will give us the rotation. 
		 * If we create two unit quaternions q1 = [mouse start vect, 0]
		 * and q2 = [mouse end vect, 0], then q2*inv(q1) converts
		 * to the arc between them to a rotation. We get:
		 * q = q2*inv(q1) = 
		 * [(mouse start vect) cross (mouse end vect), 
		 * (mouse start vect) dot (mouse end vect)]
		 */
		
		rotQuat[0] = perpVect[0];
		rotQuat[1] = perpVect[1];
		rotQuat[2] = perpVect[2];
		rotQuat[3] = dotProduct(mouseStartVect, mouseEndVect);
	} 
	
	/* Convert quaternion to rotation matrix. */
	build3x3RotMatrix(rotQuat);
	/* Multiply by old rotation. */
	matrixMultiply3(thisRotation, lastRotation);
	/* Convert to 4x4 matrix to be able to feed it to GL. */
	build4x4RotMatrix(thisRotation);	
}

void beginArcBallRotation(GLfloat mouseX, GLfloat mouseY)
{
	/* Called when the mouse button is pressed down. */
	mapToSphere(mouseX, mouseY, mouseStartVect);
}

void endArcBallRotation(void)
{
	memcpy(lastRotation, thisRotation, 9 * sizeof(GLfloat));
}

void setArcBallBounds(GLfloat width, GLfloat height)
{
	adjWidth = 1.0 / ((width - 1.0) * 0.5);
	adjHeight = 1.0 / ((height - 1.0) * 0.5);
}

void build3x3RotMatrix(GLfloat q[4])
{
	/* Converts the quaternion q to a 3x3 rotation matrix. */
	GLfloat n, s;
	GLfloat xs, ys, zs;
	GLfloat wx, wy, wz;
	GLfloat xx, xy, xz;
	GLfloat yy, yz, zz;
	
	n = q[0] * q[0] + q[1] * q[1] + q[2] * q[2] + q[3] * q[3];
	
	s = (n > 0.0) ? (2.0 / n) : 0.0;

	xs = q[0] * s;
	ys = q[1] * s;
	zs = q[2] * s;
	wx = q[3] * xs;
	wy = q[3] * ys;
	wz = q[3] * zs;
	xx = q[0] * xs;
	xy = q[0] * ys;
	xz = q[0] * zs;
	yy = q[1] * ys;
	yz = q[1] * zs;
	zz = q[2] * zs;
	
	thisRotation[0][0] = 1.0 - (yy + zz);
	thisRotation[1][0] = xy - wz;
	thisRotation[2][0] = xz + wy;

	thisRotation[0][1] = xy + wz;
	thisRotation[1][1] = 1.0 - (xx + zz);
	thisRotation[2][1] = yz - wx;

	thisRotation[0][2] = xz - wy;
	thisRotation[1][2] = yz + wx;
	thisRotation[2][2] = 1.0 - (xx + yy); 
}

void build4x4RotMatrix(GLfloat totalRot[3][3])
{
	transformMatrix[0] = totalRot[0][0];
	transformMatrix[1] = totalRot[1][0];
	transformMatrix[2] = totalRot[2][0];
	
	transformMatrix[4] = totalRot[0][1];
	transformMatrix[5] = totalRot[1][1];
	transformMatrix[6] = totalRot[2][1];
	
	transformMatrix[8] = totalRot[0][2];
	transformMatrix[9] = totalRot[1][2];
	transformMatrix[10] = totalRot[2][2];
}

GLfloat *getArcBallRotation(void)
{
	/* Returns the rotation of the arcball. */
	return transformMatrix;
}

